from contextlib import AsyncExitStack
from typing import Any, Coroutine, TypeVar

from anyio import create_task_group

from ._task import Task, _task_group, create_task as _create_task

T = TypeVar("T")


class TaskGroup:
    async def __aenter__(self) -> "TaskGroup":
        async with AsyncExitStack() as exit_stack:
            tg = await exit_stack.enter_async_context(create_task_group())
            _task_group.set(tg)
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    def create_task(self, coro: Coroutine[Any, Any, T]) -> Task[T]:
        return _create_task(coro, _task_group.get())



from anyio import Event as _Event


class Event:
    def __init__(self) -> None:
        self._event = _Event()

    async def wait(self) -> bool:
        await self._event.wait()
        return True

    def set(self) -> None:
        self._event.set()

    def is_set(self) -> bool:
        return self._event.is_set()

    def clear(self) -> None:
        self._event = _Event()



from __future__ import annotations

from typing import Any, Iterable, Literal

from anyio import create_memory_object_stream, move_on_after
from anyio.abc import TaskGroup
from anyio.streams.memory import MemoryObjectSendStream

from ._task import Task, _task_group


ALL_COMPLETED: Literal["ALL_COMPLETED"] = "ALL_COMPLETED"
FIRST_COMPLETED: Literal["FIRST_COMPLETED"] = "FIRST_COMPLETED"
FIRST_EXCEPTION: Literal["FIRST_EXCEPTION"] = "FIRST_EXCEPTION"


async def _run_and_put_task(
    task: Task,
    send_stream: MemoryObjectSendStream[Any],
):
    exc = None
    try:
        await task.wait()
    except Exception as e:
        exc = e
    try:
        await send_stream.send((task, exc))
    except Exception:
        pass


async def wait(
    aws: Iterable[Task],
    task_group: TaskGroup | None = None,
    *,
    timeout: float | int | None = None,
    return_when: Literal["ALL_COMPLETED", "FIRST_COMPLETED", "FIRST_EXCEPTION"] = ALL_COMPLETED,
):
    if task_group is None:
        task_group = _task_group.get()
    for aw in aws:
        if not isinstance(aw, Task):
            raise TypeError(f"Pass tasks, not {type(aw)}")
    if timeout is None:
        timeout = float("inf")
    done = set()
    pending = set(aws)
    send_stream, receive_stream = create_memory_object_stream[Any]()
    async with send_stream, receive_stream:
        for task in aws:
            task_group.start_soon(_run_and_put_task, task, send_stream)
        with move_on_after(timeout):
            async for aw_exc in receive_stream:
                aw, exc = aw_exc
                done.add(aw)
                pending.remove(aw)
                if return_when == FIRST_EXCEPTION and exc is not None:
                    break
                if return_when == FIRST_COMPLETED:
                    break
                if not pending:
                    break
        return done, pending



import pytest

from anyioutils import TaskGroup, create_task, wait

pytestmark = pytest.mark.anyio


async def test_task_group():
    async def coro(res):
        return res

    async def foo():
        tasks = [create_task(coro(2)), create_task(coro(3))]
        return await wait(tasks)

    async with TaskGroup() as tg:
        task0 = tg.create_task(coro(0))
        task1 = tg.create_task(coro(1))
        task2 = tg.create_task(foo())

    assert task0.result() == 0
    assert task1.result() == 1
    done, pending = task2.result()
    assert not pending
    assert sum([task.result() for task in done]) == 5
